# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import heatclient
import mock
from neutron.openstack.common import jsonutils
from neutron.openstack.common import uuidutils
from neutron.plugins.common import constants
import webob

from gbpservice.neutron.services.servicechain import config
import gbpservice.neutron.services.servicechain.drivers.simplechain_driver as\
                                                simplechain_driver
from gbpservice.neutron.tests.unit.services.servicechain import \
                                    test_servicechain_plugin

STACK_DELETE_RETRIES = 5
STACK_DELETE_RETRY_WAIT = 3


class MockStackObject(object):
    def __init__(self, status):
        self.stack_status = status


class MockHeatClientFunctions(object):
    def delete(self, stack_id):
        raise heatclient.exc.HTTPNotFound()

    def create(self, **fields):
        return {'stack': {'id': uuidutils.generate_uuid()}}

    def get(self, stack_id):
        return MockStackObject('DELETE_COMPLETE')


class MockHeatClient(object):
    def __init__(self, api_version, endpoint, **kwargs):
        self.stacks = MockHeatClientFunctions()


class SimpleChainDriverTestCase(
        test_servicechain_plugin.ServiceChainPluginTestCase):

    def setUp(self):
        config.cfg.CONF.set_override('servicechain_drivers',
                                     ['simplechain_driver'],
                                     group='servicechain')
        config.cfg.CONF.set_override('stack_delete_retries',
                                     STACK_DELETE_RETRIES,
                                     group='servicechain')
        config.cfg.CONF.set_override('stack_delete_retry_wait',
                                     STACK_DELETE_RETRY_WAIT,
                                     group='servicechain')
        super(SimpleChainDriverTestCase, self).setUp()


class TestServiceChainInstance(SimpleChainDriverTestCase):

    def test_invalid_service_type_rejected(self):
        res = self.create_servicechain_node(
                    service_type="test", config='{}',
                    expected_res_status=webob.exc.HTTPBadRequest.code)
        self.assertEqual('InvalidServiceTypeForReferenceDriver',
                         res['NeutronError']['type'])

    def test_chain_node_create_success(self):
        res = self.create_servicechain_node(
                    service_type=constants.FIREWALL, config='{}',
                    expected_res_status=webob.exc.HTTPCreated.code)
        self.assertEqual(constants.FIREWALL,
                         res['servicechain_node']['service_type'])

    def test_chain_spec_update(self):
        template1 = '{"key1":"value1"}'
        scn = self.create_servicechain_node(config=template1)
        scn1_name = scn['servicechain_node']['name']
        scn_id = scn['servicechain_node']['id']
        name = "scs1"
        template2 = '{"key2":"value2"}'
        scn2 = self.create_servicechain_node(config=template2)
        scn2_id = scn2['servicechain_node']['id']
        scn2_name = scn2['servicechain_node']['name']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        stack1 = {'stack': {'id': uuidutils.generate_uuid()}}
        stack2 = {'stack': {'id': uuidutils.generate_uuid()}}
        stack3 = {'stack': {'id': uuidutils.generate_uuid()}}
        expected_create_calls = []
        expected_delete_calls = []
        with contextlib.nested(
            mock.patch.object(simplechain_driver.HeatClient,
                              'create'),
            mock.patch.object(simplechain_driver.HeatClient,
                              'delete'),
        ) as (stack_create, stack_delete):
            stack_create.return_value = stack1
            instance1_name = "sc_instance_1"
            sc_instance1 = self.create_servicechain_instance(
                                        name=instance1_name,
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual(
                sc_instance1['servicechain_instance']['servicechain_specs'],
                [sc_spec_id])
            stack_name = "stack_" + instance1_name + scn1_name + scn_id[:5]
            expected_create_calls.append(
                        mock.call(stack_name, jsonutils.loads(template1), {}))
            stack_create.return_value = stack2
            instance2_name = "sc_instance_2"
            sc_instance2 = self.create_servicechain_instance(
                                        name=instance2_name,
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual(
                sc_instance2['servicechain_instance']['servicechain_specs'],
                [sc_spec_id])
            stack_name = "stack_" + instance2_name + scn1_name + scn_id[:5]
            expected_create_calls.append(
                        mock.call(stack_name, jsonutils.loads(template1), {}))

            #Now perform an update of the spec
            new_spec = {'servicechain_spec': {'nodes': [scn2_id]}}
            stack_create.return_value = stack3
            req = self.new_update_request(
                        'servicechain_specs', new_spec, sc_spec_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(res.status_int, webob.exc.HTTPOk.code)
            # The two existing stacks will be deleted and two new stacks
            # will be created
            expected_delete_calls.append(mock.call(stack1['stack']['id']))
            expected_delete_calls.append(mock.call(stack2['stack']['id']))
            stack_name = "stack_" + instance1_name + scn2_name + scn2_id[:5]
            expected_create_calls.append(
                        mock.call(stack_name, jsonutils.loads(template2), {}))
            stack_name = "stack_" + instance2_name + scn2_name + scn2_id[:5]
            expected_create_calls.append(
                        mock.call(stack_name, jsonutils.loads(template2), {}))
            self.assertEqual(expected_delete_calls,
                             stack_delete.call_args_list)
            self.assertEqual(expected_create_calls,
                             stack_create.call_args_list)

    def test_chain_instance_create(self):
        name = "scs1"
        scn = self.create_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        expected_stack_name = ("stack_" + "sc_instance_1" +
                               scn['servicechain_node']['name'] + scn_id[:5])
        with mock.patch.object(simplechain_driver.HeatClient,
                               'create') as stack_create:
            stack_create.return_value = {'stack': {
                                        'id': uuidutils.generate_uuid()}}
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual(
                sc_instance['servicechain_instance']['servicechain_specs'],
                [sc_spec_id])
            stack_create.assert_called_once_with(
                                    expected_stack_name, mock.ANY, mock.ANY)

    def test_chain_instance_delete(self):
        name = "scs1"
        scn = self.create_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        with mock.patch.object(simplechain_driver.HeatClient,
                               'create') as stack_create:
            stack_create.return_value = {'stack': {
                                        'id': uuidutils.generate_uuid()}}
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual(
                sc_instance['servicechain_instance']['servicechain_specs'],
                [sc_spec_id])
            with mock.patch.object(simplechain_driver.HeatClient,
                                   'delete'):
                req = self.new_delete_request(
                                    'servicechain_instances',
                                    sc_instance['servicechain_instance']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)

    def test_wait_stack_delete_for_instance_delete(self):
        name = "scs1"
        scn = self.create_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        with mock.patch.object(simplechain_driver.HeatClient,
                               'create') as stack_create:
            stack_create.return_value = {'stack': {
                                        'id': uuidutils.generate_uuid()}}
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual(
                sc_instance['servicechain_instance']['servicechain_specs'],
                [sc_spec_id])

            # Verify that as part of delete service chain instance we call
            # get method for heat stack 5 times before giving up if the state
            # does not become DELETE_COMPLETE
            with contextlib.nested(
                mock.patch.object(simplechain_driver.HeatClient, 'delete'),
                mock.patch.object(simplechain_driver.HeatClient, 'get')) as (
                                                    stack_delete, stack_get):
                stack_get.return_value = MockStackObject('PENDING_DELETE')
                req = self.new_delete_request(
                                    'servicechain_instances',
                                    sc_instance['servicechain_instance']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)
                stack_delete.assert_called_once_with(mock.ANY)
                self.assertEqual(stack_get.call_count, STACK_DELETE_RETRIES)

            # Create and delete another service chain instance and verify that
            # we call get method for heat stack only once if the stack state
            # is DELETE_COMPLETE
            sc_instance = self.create_servicechain_instance(
                                        name="sc_instance_1",
                                        servicechain_specs=[sc_spec_id])
            self.assertEqual(
                sc_instance['servicechain_instance']['servicechain_specs'],
                [sc_spec_id])
            with contextlib.nested(
                mock.patch.object(simplechain_driver.HeatClient, 'delete'),
                mock.patch.object(simplechain_driver.HeatClient, 'get')) as (
                                                    stack_delete, stack_get):
                stack_get.return_value = MockStackObject('DELETE_COMPLETE')
                req = self.new_delete_request(
                                    'servicechain_instances',
                                    sc_instance['servicechain_instance']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)
                stack_delete.assert_called_once_with(mock.ANY)
                self.assertEqual(stack_get.call_count, 1)

    def test_stack_not_found_ignored(self):
        name = "scs1"
        scn = self.create_servicechain_node()
        scn_id = scn['servicechain_node']['id']
        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])
        sc_spec_id = scs['servicechain_spec']['id']

        mock.patch(heatclient.__name__ + ".client.Client",
                   new=MockHeatClient).start()
        sc_instance = self.create_servicechain_instance(
                                    name="sc_instance_1",
                                    servicechain_specs=[sc_spec_id])
        req = self.new_delete_request(
                            'servicechain_instances',
                            sc_instance['servicechain_instance']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from neutron.api import extensions
from neutron.api.v2 import attributes
from neutron.db import model_base

from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)
from gbpservice.neutron.services.grouppolicy import config
from gbpservice.neutron.tests.unit import common as cm
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    extensions as test_ext)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_grouppolicy_plugin as test_plugin)


class ExtensionDriverTestCase(
        test_plugin.GroupPolicyPluginTestCase):

    _extension_drivers = ['test']

    def setUp(self):
        config.cfg.CONF.set_override('extension_drivers',
                                     self._extension_drivers,
                                     group='group_policy')
        super(ExtensionDriverTestCase, self).setUp()

    def test_pt_attr(self):
        # Test create with default value.
        pt = self.create_policy_target()
        pt_id = pt['policy_target']['id']
        val = pt['policy_target']['pt_extension']
        self.assertEqual("", val)
        req = self.new_show_request('policy_targets', pt_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('policy_targets')
        val = res['policy_targets'][0]['pt_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        pt = self.create_policy_target(pt_extension="abc")
        pt_id = pt['policy_target']['id']
        val = pt['policy_target']['pt_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_targets', pt_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_target': {'pt_extension': "def"}}
        req = self.new_update_request('policy_targets', data, pt_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_targets', pt_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertEqual("def", val)

    def test_ptg_attr(self):
        # Test create with default value.
        ptg = self.create_policy_target_group()
        ptg_id = ptg['policy_target_group']['id']
        val = ptg['policy_target_group']['ptg_extension']
        self.assertEqual("", val)
        req = self.new_show_request('policy_target_groups', ptg_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('policy_target_groups')
        val = res['policy_target_groups'][0]['ptg_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        ptg = self.create_policy_target_group(ptg_extension="abc")
        ptg_id = ptg['policy_target_group']['id']
        val = ptg['policy_target_group']['ptg_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_target_groups', ptg_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_target_group': {'ptg_extension': "def"}}
        req = self.new_update_request('policy_target_groups', data, ptg_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_target_groups', ptg_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertEqual("def", val)

    def test_l2p_attr(self):
        # Test create with default value.
        l2p = self.create_l2_policy()
        l2p_id = l2p['l2_policy']['id']
        val = l2p['l2_policy']['l2p_extension']
        self.assertEqual("", val)
        req = self.new_show_request('l2_policies', l2p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('l2_policies')
        val = res['l2_policies'][0]['l2p_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        l2p = self.create_l2_policy(l2p_extension="abc")
        l2p_id = l2p['l2_policy']['id']
        val = l2p['l2_policy']['l2p_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('l2_policies', l2p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'l2_policy': {'l2p_extension': "def"}}
        req = self.new_update_request('l2_policies', data, l2p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('l2_policies', l2p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertEqual("def", val)

    def test_l3p_attr(self):
        # Test create with default value.
        l3p = self.create_l3_policy()
        l3p_id = l3p['l3_policy']['id']
        val = l3p['l3_policy']['l3p_extension']
        self.assertEqual("", val)
        req = self.new_show_request('l3_policies', l3p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('l3_policies')
        val = res['l3_policies'][0]['l3p_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        l3p = self.create_l3_policy(l3p_extension="abc")
        l3p_id = l3p['l3_policy']['id']
        val = l3p['l3_policy']['l3p_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('l3_policies', l3p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'l3_policy': {'l3p_extension': "def"}}
        req = self.new_update_request('l3_policies', data, l3p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('l3_policies', l3p_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertEqual("def", val)

    def test_pc_attr(self):
        # Test create with default value.
        pc = self.create_policy_classifier()
        pc_id = pc['policy_classifier']['id']
        val = pc['policy_classifier']['pc_extension']
        self.assertEqual("", val)
        req = self.new_show_request('policy_classifiers', pc_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('policy_classifiers')
        val = res['policy_classifiers'][0]['pc_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        pc = self.create_policy_classifier(pc_extension="abc")
        pc_id = pc['policy_classifier']['id']
        val = pc['policy_classifier']['pc_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_classifiers', pc_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_classifier': {'pc_extension': "def"}}
        req = self.new_update_request('policy_classifiers', data, pc_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_classifiers', pc_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertEqual("def", val)

    def test_pa_attr(self):
        # Test create with default value.
        pa = self.create_policy_action()
        pa_id = pa['policy_action']['id']
        val = pa['policy_action']['pa_extension']
        self.assertEqual("", val)
        req = self.new_show_request('policy_actions', pa_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('policy_actions')
        val = res['policy_actions'][0]['pa_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        pa = self.create_policy_action(pa_extension="abc")
        pa_id = pa['policy_action']['id']
        val = pa['policy_action']['pa_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_actions', pa_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_action': {'pa_extension': "def"}}
        req = self.new_update_request('policy_actions', data, pa_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_actions', pa_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertEqual("def", val)

    def test_pr_attr(self):
        # Create necessary parameters.
        classifier = self.create_policy_classifier(
            name="class1", protocol="tcp", direction="out",
            port_range="50:100")
        classifier_id = classifier['policy_classifier']['id']

        # Test create with default value.
        pr = self.create_policy_rule(policy_classifier_id=classifier_id)
        pr_id = pr['policy_rule']['id']
        val = pr['policy_rule']['pr_extension']
        self.assertEqual("", val)
        req = self.new_show_request('policy_rules', pr_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('policy_rules')
        val = res['policy_rules'][0]['pr_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        pr = self.create_policy_rule(policy_classifier_id=classifier_id,
                                     pr_extension="abc")
        pr_id = pr['policy_rule']['id']
        val = pr['policy_rule']['pr_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_rules', pr_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_rule': {'pr_extension': "def"}}
        req = self.new_update_request('policy_rules', data, pr_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_rules', pr_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertEqual("def", val)

    def test_prs_attr(self):
        # Test create with default value.
        prs = self.create_policy_rule_set(policy_rules=[])
        prs_id = prs['policy_rule_set']['id']
        val = prs['policy_rule_set']['prs_extension']
        self.assertEqual("", val)
        req = self.new_show_request('policy_rule_sets', prs_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('policy_rule_sets')
        val = res['policy_rule_sets'][0]['prs_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        prs = self.create_policy_rule_set(policy_rules=[], prs_extension="abc")
        prs_id = prs['policy_rule_set']['id']
        val = prs['policy_rule_set']['prs_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_rule_sets', prs_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_rule_set': {'prs_extension': "def"}}
        req = self.new_update_request('policy_rule_sets', data, prs_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_rule_sets', prs_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertEqual("def", val)

    def test_nsp_attr(self):
        # Test create with default value.
        nsp = self.create_network_service_policy()
        nsp_id = nsp['network_service_policy']['id']
        val = nsp['network_service_policy']['nsp_extension']
        self.assertEqual("", val)
        req = self.new_show_request('network_service_policies', nsp_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list('network_service_policies')
        val = res['network_service_policies'][0]['nsp_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        nsp = self.create_network_service_policy(nsp_extension="abc")
        nsp_id = nsp['network_service_policy']['id']
        val = nsp['network_service_policy']['nsp_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('network_service_policies', nsp_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'network_service_policy': {'nsp_extension': "def"}}
        req = self.new_update_request('network_service_policies', data, nsp_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('network_service_policies', nsp_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertEqual("def", val)

    def test_es_attr(self):
        self._test_attr('external_segment')

    def test_ep_attr(self):
        self._test_attr('external_policy')

    def test_np_attr(self):
        self._test_attr('nat_pool')

    def _test_attr(self, type):
        # Test create with default value.
        acronim = _acronim(type)
        plural = cm.get_resource_plural(type)
        obj = getattr(self, 'create_%s' % type)()
        id = obj[type]['id']
        val = obj[type][acronim + '_extension']
        self.assertEqual("", val)
        req = self.new_show_request(plural, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertEqual("", val)

        # Test list.
        res = self._list(plural)
        val = res[plural][0][acronim + '_extension']
        self.assertEqual("", val)

        # Test create with explict value.
        kwargs = {acronim + '_extension': "abc"}
        obj = getattr(self, 'create_%s' % type)(**kwargs)
        id = obj[type]['id']
        val = obj[type][acronim + '_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request(plural, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {type: {acronim + '_extension': "def"}}
        req = self.new_update_request(plural, data, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertEqual("def", val)
        req = self.new_show_request(plural, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertEqual("def", val)


class TestPolicyTargetExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_target_extension'
    pt_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_policy_targets.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    value = sa.Column(sa.String(64))


class TestPolicyTargetGroupExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_target_group_extension'
    ptg_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_policy_target_groups.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    value = sa.Column(sa.String(64))


class TestL2PolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_l2_policy_extension'
    l2p_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_l2_policies.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    value = sa.Column(sa.String(64))


class TestL3PolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_l3_policy_extension'
    l3p_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_l3_policies.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    value = sa.Column(sa.String(64))


class TestPolicyClassifierExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_classifier_extension'
    pc_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_policy_classifiers.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    value = sa.Column(sa.String(64))


class TestPolicyActionExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_action_extension'
    pa_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_policy_actions.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    value = sa.Column(sa.String(64))


class TestPolicyRuleExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_rule_extension'
    pr_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_policy_rules.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    value = sa.Column(sa.String(64))


class TestPolicyRuleSetExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_rule_set_extension'
    prs_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_policy_rule_sets.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    value = sa.Column(sa.String(64))


class TestNetworkServicePolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_network_service_policy_extension'
    nsp_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_network_service_policies.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    value = sa.Column(sa.String(64))


class TestExternalSegmentExtension(model_base.BASEV2):
    __tablename__ = 'test_external_segment_extension'
    es_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_external_segments.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    value = sa.Column(sa.String(64))


class TestExternalPolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_external_policy_extension'
    ep_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_external_policies.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    value = sa.Column(sa.String(64))


class TestNatPoolExtension(model_base.BASEV2):
    __tablename__ = 'test_nat_pool_extension'
    np_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_nat_pools.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    value = sa.Column(sa.String(64))


class TestExtensionDriver(api.ExtensionDriver):
    _supported_extension_alias = 'test_extension'

    def initialize(self):
        # self.network_extension = 'Test_Network_Extension'
        # self.subnet_extension = 'Test_Subnet_Extension'
        # self.port_extension = 'Test_Port_Extension'
        extensions.append_api_extensions_path(test_ext.__path__)

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_create_policy_target(self, session, data, result):
        value = data['policy_target']['pt_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestPolicyTargetExtension(pt_id=result['id'],
                                           value=value)
        session.add(record)
        result['pt_extension'] = value

    def process_update_policy_target(self, session, data, result):
        record = (session.query(TestPolicyTargetExtension).
                  filter_by(pt_id=result['id']).
                  one())
        value = data['policy_target'].get('pt_extension')
        if value and value != record.value:
            record.value = value
        result['pt_extension'] = record.value

    def extend_policy_target_dict(self, session, result):
        record = (session.query(TestPolicyTargetExtension).
                  filter_by(pt_id=result['id']).
                  one())
        result['pt_extension'] = record.value

    def process_create_policy_target_group(self, session, data, result):
        value = data['policy_target_group']['ptg_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestPolicyTargetGroupExtension(ptg_id=result['id'],
                                                value=value)
        session.add(record)
        result['ptg_extension'] = value

    def process_update_policy_target_group(self, session, data, result):
        record = (session.query(TestPolicyTargetGroupExtension).
                  filter_by(ptg_id=result['id']).
                  one())
        value = data['policy_target_group'].get('ptg_extension')
        if value and value != record.value:
            record.value = value
        result['ptg_extension'] = record.value

    def extend_policy_target_group_dict(self, session, result):
        record = (session.query(TestPolicyTargetGroupExtension).
                  filter_by(ptg_id=result['id']).
                  one())
        result['ptg_extension'] = record.value

    def process_create_l2_policy(self, session, data, result):
        value = data['l2_policy']['l2p_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestL2PolicyExtension(l2p_id=result['id'], value=value)
        session.add(record)
        result['l2p_extension'] = value

    def process_update_l2_policy(self, session, data, result):
        record = (session.query(TestL2PolicyExtension).
                  filter_by(l2p_id=result['id']).
                  one())
        value = data['l2_policy'].get('l2p_extension')
        if value and value != record.value:
            record.value = value
        result['l2p_extension'] = record.value

    def extend_l2_policy_dict(self, session, result):
        record = (session.query(TestL2PolicyExtension).
                  filter_by(l2p_id=result['id']).
                  one())
        result['l2p_extension'] = record.value

    def process_create_l3_policy(self, session, data, result):
        value = data['l3_policy']['l3p_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestL3PolicyExtension(l3p_id=result['id'], value=value)
        session.add(record)
        result['l3p_extension'] = value

    def process_update_l3_policy(self, session, data, result):
        record = (session.query(TestL3PolicyExtension).
                  filter_by(l3p_id=result['id']).
                  one())
        value = data['l3_policy'].get('l3p_extension')
        if value and value != record.value:
            record.value = value
        result['l3p_extension'] = record.value

    def extend_l3_policy_dict(self, session, result):
        record = (session.query(TestL3PolicyExtension).
                  filter_by(l3p_id=result['id']).
                  one())
        result['l3p_extension'] = record.value

    def process_create_policy_classifier(self, session, data, result):
        value = data['policy_classifier']['pc_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestPolicyClassifierExtension(pc_id=result['id'], value=value)
        session.add(record)
        result['pc_extension'] = value

    def process_update_policy_classifier(self, session, data, result):
        record = (session.query(TestPolicyClassifierExtension).
                  filter_by(pc_id=result['id']).
                  one())
        value = data['policy_classifier'].get('pc_extension')
        if value and value != record.value:
            record.value = value
        result['pc_extension'] = record.value

    def extend_policy_classifier_dict(self, session, result):
        record = (session.query(TestPolicyClassifierExtension).
                  filter_by(pc_id=result['id']).
                  one())
        result['pc_extension'] = record.value

    def process_create_policy_action(self, session, data, result):
        value = data['policy_action']['pa_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestPolicyActionExtension(pa_id=result['id'], value=value)
        session.add(record)
        result['pa_extension'] = value

    def process_update_policy_action(self, session, data, result):
        record = (session.query(TestPolicyActionExtension).
                  filter_by(pa_id=result['id']).
                  one())
        value = data['policy_action'].get('pa_extension')
        if value and value != record.value:
            record.value = value
        result['pa_extension'] = record.value

    def extend_policy_action_dict(self, session, result):
        record = (session.query(TestPolicyActionExtension).
                  filter_by(pa_id=result['id']).
                  one())
        result['pa_extension'] = record.value

    def process_create_policy_rule(self, session, data, result):
        value = data['policy_rule']['pr_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestPolicyRuleExtension(pr_id=result['id'], value=value)
        session.add(record)
        result['pr_extension'] = value

    def process_update_policy_rule(self, session, data, result):
        record = (session.query(TestPolicyRuleExtension).
                  filter_by(pr_id=result['id']).
                  one())
        value = data['policy_rule'].get('pr_extension')
        if value and value != record.value:
            record.value = value
        result['pr_extension'] = record.value

    def extend_policy_rule_dict(self, session, result):
        record = (session.query(TestPolicyRuleExtension).
                  filter_by(pr_id=result['id']).
                  one())
        result['pr_extension'] = record.value

    def process_create_policy_rule_set(self, session, data, result):
        value = data['policy_rule_set']['prs_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestPolicyRuleSetExtension(prs_id=result['id'], value=value)
        session.add(record)
        result['prs_extension'] = value

    def process_update_policy_rule_set(self, session, data, result):
        record = (session.query(TestPolicyRuleSetExtension).
                  filter_by(prs_id=result['id']).
                  one())
        value = data['policy_rule_set'].get('prs_extension')
        if value and value != record.value:
            record.value = value
        result['prs_extension'] = record.value

    def extend_policy_rule_set_dict(self, session, result):
        record = (session.query(TestPolicyRuleSetExtension).
                  filter_by(prs_id=result['id']).
                  one())
        result['prs_extension'] = record.value

    def process_create_network_service_policy(self, session, data, result):
        value = data['network_service_policy']['nsp_extension']
        if not attributes.is_attr_set(value):
            value = ''
        record = TestNetworkServicePolicyExtension(nsp_id=result['id'],
                                                   value=value)
        session.add(record)
        result['nsp_extension'] = value

    def process_update_network_service_policy(self, session, data, result):
        record = (session.query(TestNetworkServicePolicyExtension).
                  filter_by(nsp_id=result['id']).
                  one())
        value = data['network_service_policy'].get('nsp_extension')
        if value and value != record.value:
            record.value = value
        result['nsp_extension'] = record.value

    def extend_network_service_policy_dict(self, session, result):
        record = (session.query(TestNetworkServicePolicyExtension).
                  filter_by(nsp_id=result['id']).
                  one())
        result['nsp_extension'] = record.value

    def process_create_external_segment(self, session, data, result):
        self._process_create(session, data, result, 'external_segment',
                             TestExternalSegmentExtension)

    def process_update_external_segment(self, session, data, result):
        self._process_update(session, data, result, 'external_segment',
                             TestExternalSegmentExtension)

    def extend_external_segment_dict(self, session, result):
        self._extend(session, result, 'external_segment',
                     TestExternalSegmentExtension)

    def process_create_external_policy(self, session, data, result):
        self._process_create(session, data, result, 'external_policy',
                             TestExternalPolicyExtension)

    def process_update_external_policy(self, session, data, result):
        self._process_update(session, data, result, 'external_policy',
                             TestExternalPolicyExtension)

    def extend_external_policy_dict(self, session, result):
        self._extend(session, result, 'external_policy',
                     TestExternalPolicyExtension)

    def process_create_nat_pool(self, session, data, result):
        self._process_create(session, data, result, 'nat_pool',
                             TestNatPoolExtension)

    def process_update_nat_pool(self, session, data, result):
        self._process_update(session, data, result, 'nat_pool',
                             TestNatPoolExtension)

    def extend_nat_pool_dict(self, session, result):
        self._extend(session, result, 'nat_pool', TestNatPoolExtension)

    def _process_create(self, session, data, result, type, klass):
        acronim = _acronim(type)
        value = data[type][acronim + '_extension']
        if not attributes.is_attr_set(value):
            value = ''
        kwargs = {acronim + '_id': result['id'], 'value': value}
        record = klass(**kwargs)
        session.add(record)
        result[acronim + '_extension'] = value

    def _process_update(self, session, data, result, type, klass):
        acronim = _acronim(type)
        kwargs = {acronim + '_id': result['id']}
        record = session.query(klass).filter_by(**kwargs).one()
        value = data[type].get(acronim + '_extension')
        if value and value != record.value:
            record.value = value
        result[acronim + '_extension'] = record.value

    def _extend(self, session, result, type, klass):
        acronim = _acronim(type)
        kwargs = {acronim + '_id': result['id']}
        record = session.query(klass).filter_by(**kwargs).one()
        result[acronim + '_extension'] = record.value


def _acronim(type):
    return ''.join([x[0] for x in type.split('_')])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa
from sqlalchemy import orm

from neutron.common import log
from neutron.db import model_base
from neutron.openstack.common import log as logging
from neutron.openstack.common import uuidutils

from gbpservice.neutron.db.grouppolicy import group_policy_db as gpdb


LOG = logging.getLogger(__name__)


class PolicyTargetMapping(gpdb.PolicyTarget):
    """Mapping of PolicyTarget to Neutron Port."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    # REVISIT(ivar): Set null on delete is a temporary workaround until Nova
    # bug 1158684 is fixed.
    port_id = sa.Column(sa.String(36), sa.ForeignKey('ports.id',
                                                     ondelete='SET NULL'),
                        nullable=True, unique=True)


class PTGToSubnetAssociation(model_base.BASEV2):
    """Many to many relation between PolicyTargetGroup and Subnets."""
    __tablename__ = 'gp_ptg_to_subnet_associations'
    policy_target_group_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_target_groups.id'),
        primary_key=True)
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey('subnets.id'),
                          primary_key=True)


class PolicyTargetGroupMapping(gpdb.PolicyTargetGroup):
    """Mapping of PolicyTargetGroup to set of Neutron Subnets."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    subnets = orm.relationship(PTGToSubnetAssociation,
                               cascade='all', lazy="joined")


class L2PolicyMapping(gpdb.L2Policy):
    """Mapping of L2Policy to Neutron Network."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    network_id = sa.Column(sa.String(36), sa.ForeignKey('networks.id'),
                           nullable=True, unique=True)


class L3PolicyRouterAssociation(model_base.BASEV2):
    """Models the many to many relation between L3Policies and Routers."""
    __tablename__ = 'gp_l3_policy_router_associations'
    l3_policy_id = sa.Column(sa.String(36), sa.ForeignKey('gp_l3_policies.id'),
                             primary_key=True)
    router_id = sa.Column(sa.String(36), sa.ForeignKey('routers.id'),
                          primary_key=True)


class L3PolicyMapping(gpdb.L3Policy):
    """Mapping of L3Policy to set of Neutron Routers."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    routers = orm.relationship(L3PolicyRouterAssociation,
                               cascade='all', lazy="joined")


class ExternalSegmentMapping(gpdb.ExternalSegment):
    """Mapping of L2Policy to Neutron Network."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey('subnets.id'),
                          nullable=True, unique=True)


class GroupPolicyMappingDbPlugin(gpdb.GroupPolicyDbPlugin):
    """Group Policy Mapping interface implementation using SQLAlchemy models.
    """

    def _make_policy_target_dict(self, pt, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_policy_target_dict(pt)
        res['port_id'] = pt.port_id
        return self._fields(res, fields)

    def _make_policy_target_group_dict(self, ptg, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_policy_target_group_dict(ptg)
        res['subnets'] = [subnet.subnet_id for subnet in ptg.subnets]
        return self._fields(res, fields)

    def _make_l2_policy_dict(self, l2p, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_l2_policy_dict(l2p)
        res['network_id'] = l2p.network_id
        return self._fields(res, fields)

    def _make_l3_policy_dict(self, l3p, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_l3_policy_dict(l3p)
        res['routers'] = [router.router_id for router in l3p.routers]
        return self._fields(res, fields)

    def _make_external_segment_dict(self, es, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_external_segment_dict(es)
        res['subnet_id'] = es.subnet_id
        return self._fields(res, fields)

    def _set_port_for_policy_target(self, context, pt_id, port_id):
        with context.session.begin(subtransactions=True):
            pt_db = self._get_policy_target(context, pt_id)
            pt_db.port_id = port_id

    def _add_subnet_to_policy_target_group(self, context, ptg_id, subnet_id):
        with context.session.begin(subtransactions=True):
            ptg_db = self._get_policy_target_group(context, ptg_id)
            assoc = PTGToSubnetAssociation(policy_target_group_id=ptg_id,
                                           subnet_id=subnet_id)
            ptg_db.subnets.append(assoc)
        return [subnet.subnet_id for subnet in ptg_db.subnets]

    def _set_network_for_l2_policy(self, context, l2p_id, network_id):
        with context.session.begin(subtransactions=True):
            l2p_db = self._get_l2_policy(context, l2p_id)
            l2p_db.network_id = network_id

    def _add_router_to_l3_policy(self, context, l3p_id, router_id):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            assoc = L3PolicyRouterAssociation(l3_policy_id=l3p_id,
                                              router_id=router_id)
            l3p_db.routers.append(assoc)
        return [router.router_id for router in l3p_db.routers]

    def _set_subnet_to_es(self, context, es_id, subnet_id):
        with context.session.begin(subtransactions=True):
            es_db = self._get_external_segment(context, es_id)
            es_db.subnet_id = subnet_id

    def _update_ess_for_l3p(self, context, l3p_id, ess):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            self._set_ess_for_l3p(context, l3p_db, ess)

    @log.log
    def create_policy_target(self, context, policy_target):
        pt = policy_target['policy_target']
        tenant_id = self._get_tenant_id_for_create(context, pt)
        with context.session.begin(subtransactions=True):
            pt_db = PolicyTargetMapping(id=uuidutils.generate_uuid(),
                                        tenant_id=tenant_id,
                                        name=pt['name'],
                                        description=pt['description'],
                                        policy_target_group_id=
                                        pt['policy_target_group_id'],
                                        port_id=pt['port_id'])
            context.session.add(pt_db)
        return self._make_policy_target_dict(pt_db)

    @log.log
    def create_policy_target_group(self, context, policy_target_group):
        ptg = policy_target_group['policy_target_group']
        tenant_id = self._get_tenant_id_for_create(context, ptg)
        with context.session.begin(subtransactions=True):
            ptg_db = PolicyTargetGroupMapping(id=uuidutils.generate_uuid(),
                                              tenant_id=tenant_id,
                                              name=ptg['name'],
                                              description=ptg['description'],
                                              l2_policy_id=ptg['l2_policy_id'],
                                              network_service_policy_id=
                                              ptg['network_service_policy_id'],
                                              shared=ptg.get('shared', False))
            context.session.add(ptg_db)
            if 'subnets' in ptg:
                for subnet in ptg['subnets']:
                    assoc = PTGToSubnetAssociation(
                        policy_target_group_id=ptg_db.id,
                        subnet_id=subnet
                    )
                    ptg_db.subnets.append(assoc)
            self._process_policy_rule_sets_for_ptg(context, ptg_db, ptg)
        return self._make_policy_target_group_dict(ptg_db)

    @log.log
    def update_policy_target_group(self, context, policy_target_group_id,
                                   policy_target_group):
        ptg = policy_target_group['policy_target_group']
        with context.session.begin(subtransactions=True):
            ptg_db = self._get_policy_target_group(
                context, policy_target_group_id)
            self._process_policy_rule_sets_for_ptg(context, ptg_db, ptg)
            if 'subnets' in ptg:
                # Add/remove associations for changes in subnets.
                new_subnets = set(ptg['subnets'])
                old_subnets = set(subnet.subnet_id
                                  for subnet in ptg_db.subnets)
                for subnet in new_subnets - old_subnets:
                    assoc = PTGToSubnetAssociation(
                        policy_target_group_id=policy_target_group_id,
                        subnet_id=subnet)
                    ptg_db.subnets.append(assoc)
                for subnet in old_subnets - new_subnets:
                    assoc = (
                        context.session.query(
                            PTGToSubnetAssociation).filter_by(
                                policy_target_group_id=policy_target_group_id,
                                subnet_id=subnet).one())
                    ptg_db.subnets.remove(assoc)
                    context.session.delete(assoc)
                # Don't update ptg_db.subnets with subnet IDs.
                del ptg['subnets']
            ptg_db.update(ptg)
        return self._make_policy_target_group_dict(ptg_db)

    @log.log
    def create_l2_policy(self, context, l2_policy):
        l2p = l2_policy['l2_policy']
        tenant_id = self._get_tenant_id_for_create(context, l2p)
        with context.session.begin(subtransactions=True):
            l2p_db = L2PolicyMapping(id=uuidutils.generate_uuid(),
                                     tenant_id=tenant_id,
                                     name=l2p['name'],
                                     description=l2p['description'],
                                     l3_policy_id=l2p['l3_policy_id'],
                                     network_id=l2p['network_id'],
                                     shared=l2p.get('shared', False))
            context.session.add(l2p_db)
        return self._make_l2_policy_dict(l2p_db)

    @log.log
    def create_l3_policy(self, context, l3_policy):
        l3p = l3_policy['l3_policy']
        self.validate_ip_pool(l3p.get('ip_pool', None), l3p['ip_version'])
        tenant_id = self._get_tenant_id_for_create(context, l3p)
        self.validate_subnet_prefix_length(l3p['ip_version'],
                                           l3p['subnet_prefix_length'],
                                           l3p.get('ip_pool', None))
        with context.session.begin(subtransactions=True):
            l3p_db = L3PolicyMapping(id=uuidutils.generate_uuid(),
                                     tenant_id=tenant_id,
                                     name=l3p['name'],
                                     ip_version=l3p['ip_version'],
                                     ip_pool=l3p['ip_pool'],
                                     subnet_prefix_length=
                                     l3p['subnet_prefix_length'],
                                     description=l3p['description'],
                                     shared=l3p.get('shared', False))
            if 'routers' in l3p:
                for router in l3p['routers']:
                    assoc = L3PolicyRouterAssociation(
                        l3_policy_id=l3p_db.id,
                        router_id=router
                    )
                    l3p_db.routers.append(assoc)
            if 'external_segments' in l3p:
                self._set_ess_for_l3p(context, l3p_db,
                                      l3p['external_segments'])
            context.session.add(l3p_db)
        return self._make_l3_policy_dict(l3p_db)

    @log.log
    def update_l3_policy(self, context, l3_policy_id, l3_policy):
        l3p = l3_policy['l3_policy']
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3_policy_id)
            if 'subnet_prefix_length' in l3p:
                self.validate_subnet_prefix_length(l3p_db.ip_version,
                                                   l3p['subnet_prefix_length'],
                                                   l3p.get('ip_pool', None))
            if 'routers' in l3p:
                # Add/remove associations for changes in routers.
                new_routers = set(l3p['routers'])
                old_routers = set(router.router_id
                                  for router in l3p_db.routers)
                for router in new_routers - old_routers:
                    assoc = L3PolicyRouterAssociation(
                        l3_policy_id=l3_policy_id, router_id=router)
                    l3p_db.routers.append(assoc)
                for router in old_routers - new_routers:
                    assoc = (context.session.query(L3PolicyRouterAssociation).
                             filter_by(l3_policy_id=l3_policy_id,
                                       router_id=router).
                             one())
                    l3p_db.routers.remove(assoc)
                    context.session.delete(assoc)
                # Don't update l3p_db.routers with router IDs.
                del l3p['routers']
            if 'external_segments' in l3p:
                self._set_ess_for_l3p(context, l3p_db,
                                      l3p['external_segments'])
                del l3p['external_segments']
            l3p_db.update(l3p)
        return self._make_l3_policy_dict(l3p_db)

    @log.log
    def create_external_segment(self, context, external_segment):
        es = external_segment['external_segment']
        tenant_id = self._get_tenant_id_for_create(context, es)
        with context.session.begin(subtransactions=True):
            es_db = ExternalSegmentMapping(
                id=uuidutils.generate_uuid(), tenant_id=tenant_id,
                name=es['name'], description=es['description'],
                shared=es.get('shared', False), ip_version=es['ip_version'],
                cidr=es['cidr'],
                port_address_translation=es['port_address_translation'],
                subnet_id=es['subnet_id'])
            context.session.add(es_db)
            if 'external_routes' in es:
                self._process_segment_ers(context, es_db, es)
        return self._make_external_segment_dict(es_db)



# Copyright 2014 Alcatel-Lucent USA Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo.config import cfg


restproxy_opts = [
    cfg.StrOpt('server', default='localhost:8800',
               help=_("IP Address and Port of Nuage's VSD server")),
    cfg.StrOpt('serverauth', default='username:password',
               secret=True,
               help=_("Username and password for authentication")),
    cfg.BoolOpt('serverssl', default=False,
                help=_("Boolean for SSL connection with VSD server")),
    cfg.StrOpt('base_uri', default='/',
               help=_("Nuage provided base uri to reach out to VSD")),
    cfg.StrOpt('organization', default='system',
               help=_("Organization name in which VSD will orchestrate "
                      "network resources using openstack")),
    cfg.StrOpt('auth_resource', default='',
               help=_("Nuage provided uri for initial authorization to "
                      "access VSD")),
    cfg.StrOpt('application', default = 'gbp_application',
               help=_("Application name in which group based policy "
                      "is going to get defined"))
]


def nuage_register_cfg_opts():
    cfg.CONF.register_opts(restproxy_opts, "RESTPROXY")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo.config import cfg


od_driver_opts = [
    cfg.StrOpt('odl_username',
               default='admin',
               help=_("OpenDaylight Controller Username")),
    cfg.StrOpt('odl_password',
               default='admin',
               help=_("OpenDaylight Controller Password")),
    cfg.StrOpt('odl_host',
               default='127.0.0.1',
               help=_("OpenDaylight Controller host ip address")),
    cfg.StrOpt('odl_port',
               default='8080',
               help=_("OpenDaylight Controller Rest API port number")),
]


cfg.CONF.register_opts(od_driver_opts, "odl_driver")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.neutron.services.grouppolicy.drivers.odl import odl_mapping
from neutron.common import constants as n_constants
from neutron.extensions import portbindings
from neutron.openstack.common import log
from neutron.plugins.common import constants
from neutron.plugins.ml2 import driver_api as api

LOG = log.getLogger(__name__)


class OdlMechanismGBPDriver(api.MechanismDriver):

    def initialize(self):
        self._odl_gbp = None
        self.vif_type = portbindings.VIF_TYPE_OVS
        self.vif_details = {portbindings.CAP_PORT_FILTER: True}

    @property
    def odl_gbp(self):
        if not self._odl_gbp:
            self._odl_gbp = (odl_mapping.OdlMappingDriver.
                             get_initialized_instance())
        return self._odl_gbp

    def create_port_postcommit(self, context):
        # DHCP Ports are created implicitly by Neutron, need to inform GBP
        if (context.current.get('device_owner') ==
                n_constants.DEVICE_OWNER_DHCP):
            self.odl_gbp.create_dhcp_policy_target_if_needed(
                context._plugin_context, context.current
            )

    def bind_port(self, context):
        LOG.debug("Attempting to bind port %(port)s on "
                  "network %(network)s",
                  {'port': context.current['id'],
                   'network': context.network.current['id']})
        for segment in context.network.network_segments:
            if self._check_segment(segment):
                context.set_binding(segment[api.ID],
                                    self.vif_type,
                                    self.vif_details,
                                    status=n_constants.PORT_STATUS_ACTIVE)
                LOG.debug("Bound using segment: %s", segment)
                return
            else:
                LOG.debug("Refusing to bind port for segment ID %(id)s, "
                          "segment %(seg)s, phys net %(physnet)s, and "
                          "network type %(nettype)s",
                          {'id': segment[api.ID],
                           'seg': segment[api.SEGMENTATION_ID],
                           'physnet': segment[api.PHYSICAL_NETWORK],
                           'nettype': segment[api.NETWORK_TYPE]})

    def _check_segment(self, segment):
        """Verify a segment is valid for the OpenDaylight MechanismDriver.

        Verify the requested segment is supported by ODL and return True or
        False to indicate this to callers.
        """
        network_type = segment[api.NETWORK_TYPE]
        return network_type in [constants.TYPE_VXLAN, ]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from neutron import manager
from neutron.openstack.common import lockutils  # noqa
from neutron.openstack.common import log as logging
from neutron.plugins.common import constants

from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db as gpdb
from gbpservice.neutron.services.grouppolicy.common import constants as g_const
from gbpservice.neutron.services.grouppolicy.common import exceptions as gpexc
from gbpservice.neutron.services.grouppolicy.drivers import (
    resource_mapping as api)
from gbpservice.neutron.services.grouppolicy.drivers.odl import odl_manager


LOG = logging.getLogger(__name__)


class ExternalSegmentNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("External Segment currently not supported on ODL GBP "
                "driver.")


class UpdateL3PolicyNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Update L3 Policy currently not supported on ODL GBP "
                "driver.")


class UpdateL2PolicyNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Update L2 Policy currently not supported on ODL GBP "
                "driver.")


class UpdatePTNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Update Policy Target currently not supported on ODL GBP "
                "driver.")


class UpdatePTGNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Update Policy Target Group currently not supported on ODL "
                "GBP driver.")


class L2PolicyMultiplePolicyTargetGroupNotSupportedOnOdlDriver(
        gpexc.GroupPolicyBadRequest):
    message = _("An L2 policy can't have multiple policy target groups on "
                "ODL GBP driver.")


class UpdatePolicyActionNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Update Policy Action currently not supported on ODL GBP "
                "driver.")


class RedirectActionNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Redirect action is currently not supported for ODL GBP "
                "driver.")


class OnlyAllowActionSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Currently only allow action is supported for ODL GBP "
                "driver.")


class UpdateClassifierNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Update Policy Classifier currently not supported on ODL GBP "
                "driver.")


class PolicyRuleUpdateNotSupportedOnOdlDriver(gpexc.GroupPolicyBadRequest):
    message = _("Policy rule update is not supported on for ODL GBP"
                "driver.")


class ExactlyOneActionPerRuleIsSupportedOnOdlDriver(
        gpexc.GroupPolicyBadRequest):
    message = _("Exactly one action per rule is supported on ODL GBP driver.")


class ClassifierTcpUdpPortRangeNotSupportedOnOdlDriver(
        gpexc.GroupPolicyBadRequest):
    message = _("Tcp or Udp port range is not supported on ODL GBP driver.")


class ClassifierUnknownIPProtocolNotSupportedOnOdlDriver(
        gpexc.GroupPolicyBadRequest):
    message = _("Unknown IP Protocol is not supported on ODL GBP driver.")


class OdlMappingDriver(api.ResourceMappingDriver):
    """ODL Mapping driver for Group Policy plugin.

    This driver implements group policy semantics by mapping group
    policy resources to various other neutron resources, and leverages
    ODL backend for enforcing the policies.
    """

    me = None
    manager = None

    @staticmethod
    def get_odl_manager():
        if not OdlMappingDriver.manager:
            OdlMappingDriver.manager = odl_manager.OdlManager()
        return OdlMappingDriver.manager

    def initialize(self):
        super(OdlMappingDriver, self).initialize()
        self.odl_manager = OdlMappingDriver.get_odl_manager()
        self._gbp_plugin = None
        OdlMappingDriver.me = self

    @property
    def gbp_plugin(self):
        if not self._gbp_plugin:
            self._gbp_plugin = (manager.NeutronManager.get_service_plugins()
                                .get("GROUP_POLICY"))
        return self._gbp_plugin

    @staticmethod
    def get_initialized_instance():
        return OdlMappingDriver.me

    def create_dhcp_policy_target_if_needed(self, plugin_context, port):
        session = plugin_context.session
        if (self._port_is_owned(session, port['id'])):
            # Nothing to do
            return

        # Retrieve PTG
        # TODO(ywu): optimize later
        subnets = self._core_plugin._get_subnets_by_network(
            plugin_context, port['network_id']
        )
        ptg = (plugin_context.session.query(gpdb.PolicyTargetGroupMapping).
               join(gpdb.PolicyTargetGroupMapping.subnets).
               filter(gpdb.PTGToSubnetAssociation.subnet_id ==
                      subnets[0]['id']).
               first())

        # Create PolicyTarget
        attrs = {'policy_target':
                 {'tenant_id': port['tenant_id'],
                  'name': 'dhcp-%s' % ptg['id'],
                  'description': ("Implicitly created DHCP policy "
                                  "target"),
                  'policy_target_group_id': ptg['id'],
                  'port_id': port['id']}}
        self.gbp_plugin.create_policy_target(plugin_context, attrs)
        # TODO(ODL): security group is not required
        # sg_id = self._ensure_default_security_group(plugin_context,
        #                                            port['tenant_id'])
        # data = {'port': {'security_groups': [sg_id]}}
        # self._core_plugin.update_port(plugin_context, port['id'], data)

    def create_external_segment_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def update_external_segment_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def delete_external_segment_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def create_external_policy_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def update_external_policy_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def delete_external_policy_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def create_nat_pool_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def update_nat_pool_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def delete_nat_pool_precommit(self, context):
        raise ExternalSegmentNotSupportedOnOdlDriver()

    def create_policy_target_postcommit(self, context):
        super(OdlMappingDriver, self).create_policy_target_postcommit(context)
        pt = self._get_pt_detail(context)
        ep = {
            "endpoint-group": pt['ptg_id'],
            "l2-context": pt['l2ctx_id'],
            "l3-address": pt['l3_list'],
            "mac-address": pt['mac_address'],
            "port-name": pt['neutron_port_id'],
            "tenant": pt['tenant_id']
        }
        self.odl_manager.register_endpoints([ep])

    def update_policy_target_precommit(self, context):
        raise UpdatePTNotSupportedOnOdlDriver()

    def delete_policy_target_postcommit(self, context):
        pt = self._get_pt_detail(context)
        ep = {
            "l2": pt['l2_list'],
            "l3": pt['l3_list']
        }
        self.odl_manager.unregister_endpoints([ep])
        # Delete Neutron's port
        super(OdlMappingDriver, self).delete_policy_target_postcommit(context)

    def create_l3_policy_postcommit(self, context):
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]
        l3ctx = {
            "id": context.current['id'],
            "name": context.current['name'],
            "description": context.current['description']
        }
        self.odl_manager.create_update_l3_context(tenant_id, l3ctx)

    def update_l3_policy_precommit(self, context):
        raise UpdateL3PolicyNotSupportedOnOdlDriver()

    def delete_l3_policy_postcommit(self, context):
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]
        l3ctx = {
            "id": context.current['id']
        }
        self.odl_manager.delete_l3_context(tenant_id, l3ctx)

    def create_l2_policy_postcommit(self, context):
        super(OdlMappingDriver, self).create_l2_policy_postcommit(context)
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]

        # l2_policy mapped to l2_bridge_domain in ODL
        l2bd = {
            "id": context.current['id'],
            "name": context.current['name'],
            "description": context.current['description'],
            "parent": context.current['l3_policy_id']
        }
        self.odl_manager.create_update_l2_bridge_domain(tenant_id, l2bd)

        # Implicit network within l2 policy mapped to l2 FD in ODL
        net_id = context.current['network_id']
        network = self._core_plugin.get_network(context._plugin_context,
                                                net_id)
        l2fd = {
            "id": net_id,
            "name": network['name'],
            "parent": context.current['id']
        }
        self.odl_manager.create_update_l2_flood_domain(tenant_id, l2fd)

    def update_l2_policy_precommit(self, context):
        raise UpdateL2PolicyNotSupportedOnOdlDriver()

    def delete_l2_policy_postcommit(self, context):
        super(OdlMappingDriver, self).delete_l2_policy_postcommit(context)
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]

        # l2_policy mapped to l2_bridge_domain in ODL
        l2bd = {
            "id": context.current['id']
        }
        self.odl_manager.delete_l2_bridge_domain(tenant_id, l2bd)

        # Implicit network within l2 policy mapped to l2 FD in ODL
        net_id = context.current['network_id']
        l2fd = {
            "id": net_id,
        }
        self.odl_manager.delete_l2_flood_domain(tenant_id, l2fd)

    def create_policy_target_group_postcommit(self, context):
        super(OdlMappingDriver, self).create_policy_target_group_postcommit(
            context)
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]
        subnets = context.current['subnets']
        provided_contract = self._make_odl_contract_and_clause(
            context, context.current['provided_policy_rule_sets']
        )
        consumed_contract = self._make_odl_contract_and_clause(
            context, context.current['consumed_policy_rule_sets']
        )

        # PTG mapped to EPG in ODL
        # TODO(ODL); add back description field after PoC
        epg = {
            "id": context.current['id'],
            "name": context.current['name'],
            "network-domain": subnets[0]
        }

        if provided_contract:
            epg['provider-named-selector'] = {
                "name": 'Contract-' + provided_contract['id'],
                "contract": provided_contract['id']
            }
            self.odl_manager.create_update_contract(tenant_id,
                                                    provided_contract)
        if consumed_contract:
            epg['consumer-named-selector'] = {
                "name": 'Contract-' + consumed_contract['id'],
                "contract": consumed_contract['id']
            }
            self.odl_manager.create_update_contract(tenant_id,
                                                    consumed_contract)

        self.odl_manager.create_update_endpoint_group(tenant_id, epg)

        # Implicit subnet within policy target group mapped to subnet in ODL
        for subnet_id in subnets:
            neutron_subnet = self._core_plugin.get_subnet(
                context._plugin_context, subnet_id
            )
            odl_subnet = {
                "id": subnet_id,
                "ip-prefix": neutron_subnet['cidr'],
                "parent": neutron_subnet['network_id'],
                "virtual-router-ip": neutron_subnet['gateway_ip']
            }
            self.odl_manager.create_update_subnet(tenant_id, odl_subnet)

    def _make_odl_contract_and_clause(self, context, rule_sets):
        # As no contract/clause in O.S., they will be generated dynamically
        # when rule sets are associated with PTG:
        # 1. an association is mapped to a contract with single clause
        # 2. rule sets mapped to subjects
        # 3. clause name is concatenation of sorted subject names
        # 4. contract ID is generated based on the clause name
        # 5. As a combination of same rule sets produce same sorted subject
        #    names, consistent clause name and contract ID are guaranteed
        contract = None
        if rule_sets:
            subjects = []
            subject_names = []
            for rule_set_id in rule_sets:
                # a subject is mapped to a rule set
                subject = self._make_subject(context, rule_set_id)
                subjects.append(subject)
                subject_names.append(subject['name'])
            clause_name = "-".join(sorted(subject_names)).encode('ascii',
                                                                 'ignore')
            contract_id = uuid.uuid3(uuid.NAMESPACE_DNS, clause_name).urn[9:]
            clauses = [
                {
                    "name": clause_name,
                    "subject-refs": subject_names
                }
            ]
            contract = {
                "id": contract_id,
                "clause": clauses,
                "subject": subjects
            }
        return contract

    def _make_subject(self, context, rule_set_id):
        rule_set = context._plugin.get_policy_rule_set(
            context._plugin_context, rule_set_id
        )
        rules = []
        for rule_id in rule_set['policy_rules']:
            rule = self._make_odl_rule(context, rule_id)
            rules.append(rule)
        return {
            "name": rule_set['name'],
            "rule": rules
        }

    def _make_odl_rule(self, context, rule_id):
        rule = context._plugin.get_policy_rule(
            context._plugin_context, rule_id
        )
        stack_classifier = context._plugin.get_policy_classifier(
            context._plugin_context, rule['policy_classifier_id']
        )

        # while openstack supports only one classifier per rule, the classifier
        # may mapped to multi classifier in ODL
        classifier_refs = []
        classifiers = self._make_odl_classifiers(stack_classifier)
        for classifier in classifiers:
            classifier_ref = {
                "name": classifier['name']
            }
            if classifier['direction'] != "bidirectional":
                classifier_ref['direction'] = classifier['direction']
            classifier_refs.append(classifier_ref)
        action_refs = []
        for action_id in rule['policy_actions']:
            action = context._plugin.get_policy_action(
                context._plugin_context, action_id
            )
            action_refs.append(
                {
                    "name": action['name']
                }
            )

        # TODO(ODL): send action_refs later but not for PoC
        return {
            "name": rule['name'],
            "classifier-ref": classifier_refs,
        }

    def update_policy_target_group_precommit(self, context):
        raise UpdatePTGNotSupportedOnOdlDriver()

    def delete_policy_target_group_postcommit(self, context):
        # TODO(ODL): delete contract if no one uses it
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]
        subnets = context.current['subnets']

        # delete mapped subnets in ODL, and clean them up from neutron
        for subnet_id in subnets:
            self._cleanup_subnet(context._plugin_context, subnet_id, None)
            odl_subnet = {
                "id": subnet_id
            }
            self.odl_manager.delete_subnet(tenant_id, odl_subnet)

        # delete mapped EPG in ODL
        epg = {
            "id": context.current['id'],
        }
        self.odl_manager.delete_endpoint_group(tenant_id, epg)

    def create_policy_action_precommit(self, context):
        # TODO(odl): allow redirect for service chaining
        if context.current['action_type'] == g_const.GP_ACTION_REDIRECT:
            raise RedirectActionNotSupportedOnOdlDriver()

    def create_policy_action_postcommit(self, context):
        super(OdlMappingDriver, self).create_policy_action_postcommit(context)
        # TODO(ODL): remove comment out after PoC
        # tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]

        # fill in action instance data
        if context.current['action_type'] == g_const.GP_ACTION_ALLOW:
            # TODO(ODL): remove the return and comment out after POC
            return
            # action_definition_id = "f942e8fd-e957-42b7-bd18-f73d11266d17"
            # action_instance = {
            #     "action-definition-id": action_definition_id,
            #     "name": context.current['name'],
            #     "parameter-value": [
            #         {
            #             "name": context.current['name'],
            #             "string-value": context.current['action_type'],
            #         }
            #     ]
            # }
            # self.odl_manager.create_action(tenant_id, action_instance)
        else:
            raise OnlyAllowActionSupportedOnOdlDriver()

    def update_policy_action_precommit(self, context):
        raise UpdatePolicyActionNotSupportedOnOdlDriver()

    def delete_policy_action_postcommit(self, context):
        super(OdlMappingDriver, self).delete_policy_action_postcommit(context)
        # TODO(ODL): remove comment out after PoC
        # tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]
        #
        # # fill in action instance data
        # action_instance = {
        #     "name": context.current['name']
        # }
        # self.odl_manager.delete_action(tenant_id, action_instance)

    def create_policy_classifier_postcommit(self, context):
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]
        classifiers = self._make_odl_classifiers(context.current)

        for classifier in classifiers:
            classifier_instance = {
                "classifier-definition-id":
                    classifier['classifier-definition-id'],
                "name": classifier['name'],
                "parameter-value": classifier['parameter-value']
            }
            self.odl_manager.create_classifier(tenant_id, classifier_instance)

    def _make_odl_classifiers(self, stack_classifier):
        classifiers = []
        if stack_classifier['protocol'] == constants.ICMP:
            direction = stack_classifier['direction']
            if direction == 'bi':
                direction = "bidirectional"
            classifier = {
                # Use hard coded value based on current ODL implementation
                "classifier-definition-id":
                    '79c6fdb2-1e1a-4832-af57-c65baf5c2335',
                "name": stack_classifier['name'],
                "parameter-value": [
                    {
                        "name": "proto",
                        # TODO(yapeng): change the hard code value
                        "int-value": 1,
                    }
                ],
                "direction": direction
            }
            classifiers.append(classifier)
        else:
            # For TCP and UDP protoocol create two classifier (in and out)
            for port in ['sourceport', 'destport']:
                if stack_classifier['direction'] == 'in':
                    if port == 'destport':
                        direction = 'in'
                    else:
                        direction = 'out'
                elif stack_classifier['direction'] == 'out':
                    if port == 'destport':
                        direction = 'out'
                    else:
                        direction = 'in'
                else:
                    direction = 'bidirectional'

                classifier = {
                    # Use hard coded value based on current ODL implementation
                    "classifier-definition-id":
                        '4250ab32-e8b8-445a-aebb-e1bd2cdd291f',
                    "direction": direction,
                    "name": stack_classifier['name'] + '-' + port,
                    "parameter-value": [
                        {
                            "name": "type",
                            "string-value": stack_classifier['protocol'],
                        },
                        {
                            "name": port,
                            "int-value": stack_classifier['port_range'],
                        }
                    ]
                }
                classifiers.append(classifier)
        return classifiers

    def update_policy_classifier_precommit(self, context):
        raise UpdateClassifierNotSupportedOnOdlDriver()

    def delete_policy_classifier_postcommit(self, context):
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]

        if context.current['protocol'] == constants.ICMP:
            # fill in classifier instance data
            classifier_instance = {
                "name": context.current['name']
            }
            self.odl_manager.delete_classifier(tenant_id, classifier_instance)
            return

        # fill in classifier instance data
        for port in ['sourceport', 'destport']:
            classifier_instance = {
                "name": context.current['name'] + '-' + port,
            }
            self.odl_manager.delete_classifier(tenant_id, classifier_instance)

    def create_policy_rule_precommit(self, context):
        if ('policy_actions' in context.current and
                len(context.current['policy_actions']) != 1):
            # TODO(odl): to be fixed when redirect is supported
            raise ExactlyOneActionPerRuleIsSupportedOnOdlDriver()

    def update_policy_rule_precommit(self, context):
        # TODO(ivar): add support for action update on policy rules
        raise PolicyRuleUpdateNotSupportedOnOdlDriver()

    def _get_pt_detail(self, context):
        port_id = context.current['port_id']
        port = self._core_plugin.get_port(context._plugin_context, port_id)
        tenant_id = uuid.UUID(context.current['tenant_id']).urn[9:]
        ptg_id = context.current['policy_target_group_id']
        ptg = self.gbp_plugin.get_policy_target_group(context._plugin_context,
                                                      ptg_id)
        l2ctx_id = ptg['l2_policy_id']
        l2ctx = self.gbp_plugin.get_l2_policy(context._plugin_context,
                                              l2ctx_id)
        l3ctx_id = l2ctx['l3_policy_id']
        mac_address = port['mac_address']
        neutron_port_id = 'tap' + port_id[:11]

        l3_list = []
        for fixed_ip in port['fixed_ips']:
            l3_list.append(
                {
                    "ip-address": fixed_ip['ip_address'],
                    "l3-context": l3ctx_id
                }
            )

        l2_list = [
            {
                "l2-context": l2ctx_id,
                "mac-address": mac_address
            }
        ]

        return {
            "port_id": port_id,
            "tenant_id": tenant_id,
            "ptg_id": ptg_id,
            "l2ctx_id": l2ctx_id,
            "l3ctx_id": l3ctx_id,
            "mac_address": mac_address,
            "neutron_port_id": neutron_port_id,
            "l3_list": l3_list,
            "l2_list": l2_list,
        }
